@@ -123,6 +123,132 @@ authldap_free_config removed - no longer required.
 #define LDAP_OPT_SUCCESS LDAP_SUCCESS
 #endif
 
+static char **l_get_values(LDAP *ld, LDAPMessage *entry, const char *attribut)
+{
+	struct berval **p=ldap_get_values_len(ld, entry, attribut);
+	int n;
+	char **a;
+
+	if (!p)
+		return NULL;
+
+	n=ldap_count_values_len(p);
+
+
+	a=malloc((n + 1) * sizeof(char *));
+
+	if (!a)
+	{
+		DPRINTF("malloc failed");
+		ldap_value_free_len(p);
+		return NULL;
+	}
+
+	int i;
+
+	for (i=0; i<n; i++)
+	{
+		if ((a[i]=malloc(p[i]->bv_len+1)) == NULL)
+		{
+			DPRINTF("malloc failed");
+			while (i--)
+			{
+				free(a[i]);
+			}
+			free(a);
+			ldap_value_free_len(p);
+			return NULL;
+		}
+
+		memcpy(a[i], p[i]->bv_val, p[i]->bv_len);
+		a[i][p[i]->bv_len]=0;
+	}
+
+	ldap_value_free_len(p);
+	a[i]=NULL;
+	return a;
+}
+
+static void l_value_free(char **p)
+{
+	int i;
+
+	for (i=0; p[i]; ++i)
+		free(p[i]);
+	free(p);
+}
+
+static int l_count_values(char **p)
+{
+	int i;
+
+	for (i=0; p[i]; ++i)
+		;
+	return i;
+}
+
+static int l_unbind(LDAP *ld)
+{
+	return ldap_unbind_ext(ld, NULL, NULL);
+}
+
+static int l_simple_bind_s(LDAP *ld,
+			   const char *who,
+			   const char *passwd)
+{
+	struct berval cred;
+
+	cred.bv_len=passwd ? strlen(passwd):0;
+	cred.bv_val=(char *)passwd;
+
+	return ldap_sasl_bind_s(ld, who, NULL, &cred, NULL, NULL, NULL);
+}
+
+static int l_search_st(LDAP *ld,
+		       const char *base,
+		       int scope,
+		       const char *filter,
+		       char **attrs,
+		       int attrsonly,
+		       struct timeval *timeout,
+		       LDAPMessage **res)
+{
+	return ldap_search_ext_s(ld, base, scope, filter, attrs,
+				 attrsonly,
+				 NULL, NULL,
+				 timeout,
+				 100,
+				 res);
+}
+
+static int l_modify_s(LDAP *ld,
+		      const char *dn,
+		      LDAPMod **mods)
+{
+	return ldap_modify_ext_s(ld, dn, mods, NULL, NULL);
+}
+
+static int l_search(LDAP *ld,
+		    const char *base,
+		    int scope,
+		    const char *filter,
+		    char **attrs,
+		    int attrsonly)
+{
+	struct timeval tv;
+	int msgid;
+
+	tv.tv_sec=60*60;
+	tv.tv_usec=0;
+
+	if (ldap_search_ext(ld, base, scope, filter, attrs, attrsonly,
+			    NULL, NULL, &tv, 1000000, &msgid) !=
+	    LDAP_SUCCESS)
+		return -1;
+
+	return msgid;
+}
+
 struct ldap_info
 {
 	const char *uri;
@@ -335,7 +461,7 @@ static int authldap_read_config(struct ldap_info *ldap)
 	memset(ldap,0,sizeof(struct ldap_info));
 	
 	if (!read_env("LDAP_URI",&ldap->uri,
-		      "You need to specify a ldap server in config file",1,NULL))
+		      "You need to specify LDAP_URI in config file",1,NULL))
 		return 0;
 
 	if (!read_env("LDAP_AUTHBIND", &p, "", 0, ""))
@@ -599,25 +725,66 @@ static int authldap_read_config(struct ldap_info *ldap)
 	return 1;
 }
 
+static void get_error(LDAP *ld, LDAPMessage *entry,
+		      const char *func,
+		      const char *attribut)
+{
+#if HAVE_LDAP_PARSE_RESULT
+
+	int errcode;
+	char *nmatched;
+	char *errmsg;
+		
+
+	if (ldap_parse_result(ld, entry, &errcode, &nmatched,
+			      &errmsg, NULL, NULL, 0)
+	    != LDAP_SUCCESS)
+	{
+		DPRINTF("ldap_parseresult failed");
+	}
+	else
+	{
+		if (errcode && errcode != LDAP_DECODING_ERROR &&
+		    errcode != LDAP_NO_RESULTS_RETURNED)
+		{
+			DPRINTF("get_values attribute %s: %s",
+				attribut,
+				errmsg ? errmsg:"unknown error");
+		}
+
+		if (errmsg)
+			ldap_memfree(errmsg);
+		if (nmatched)
+			ldap_memfree(nmatched);
+	}
+#else
+#if HAVE_LDAP_RESULT2ERROR
+	int ld_errno = ldap_result2error(ld,entry,0);
+	if (ld_errno && ld_errno != LDAP_DECODING_ERROR
+	    && ld_errno != LDAP_NO_RESULTS_RETURNED)
+	{
+		DPRINTF("get_values attribute %s: %s", attribut,
+			ldap_err2string(ld_errno));
+	}
+#else
+	if (ld->ld_errno != LDAP_DECODING_ERROR
+	    && ld->ld_errno != LDAP_NO_RESULTS_RETURNED)
+	{
+		DPRINTF("get_values attribute %s: %s", attribut,
+			ldap_err2string(ld->ld_errno));
+	}
+#endif
+#endif
+}
+
 static char **get_values(LDAP *ld, LDAPMessage *entry, const char *attribut)
 {
 	char ** values;
-	values=ldap_get_values(ld,entry, (char *)attribut);
+	values=l_get_values(ld,entry, (char *)attribut);
 
 	if (values==NULL)
 	{
-#ifdef HAVE_LDAP_RESULT2ERROR
-		int ld_errno = ldap_result2error(ld,entry,0);
-		if (ld_errno && ld_errno != LDAP_DECODING_ERROR
-		&& ld_errno != LDAP_NO_RESULTS_RETURNED)
-			DPRINTF("get_values attribute %s: %s", attribut,
-				ldap_err2string(ld_errno));
-#else
-		if (ld->ld_errno != LDAP_DECODING_ERROR
-		&& ld->ld_errno != LDAP_NO_RESULTS_RETURNED)
-			DPRINTF("get_values attribute %s: %s", attribut,
-				ldap_err2string(ld->ld_errno));
-#endif
+		get_error(ld, entry, "get_values", attribut);
 	}
 
 	return values;
@@ -637,30 +804,19 @@ static char **get_values(LDAP *ld, LDAPMessage *entry, const char *attribut)
  *   void
  */
 static void copy_value(LDAP *ld, LDAPMessage *entry, const char *attribut,
-	char **copy, const char *username)
+		       char **copy, const char *username)
 {
-  char ** values;
-  values=ldap_get_values(ld,entry, (char *)attribut);
+	char ** values;
+	values=l_get_values(ld,entry, attribut);
 
 	if (values==NULL)
 	{
-#ifdef HAVE_LDAP_RESULT2ERROR
-		int ld_errno = ldap_result2error(ld,entry,0);
-		if (ld_errno && ld_errno != LDAP_DECODING_ERROR
-		&& ld_errno != LDAP_NO_RESULTS_RETURNED)
-			DPRINTF("copy_value attribute %s for %s: %s", attribut,
-				username, ldap_err2string(ld_errno));
-#else
-		if (ld->ld_errno != LDAP_DECODING_ERROR
-		&& ld->ld_errno != LDAP_NO_RESULTS_RETURNED)
-			DPRINTF("copy_value attribute %s for %s: %s", attribut,
-				username, ldap_err2string(ld->ld_errno));
-#endif
+		get_error(ld, entry, "copy_value ", attribut);
 		*copy=NULL;
 		return;
 	}
   /* We accept only attribute with one value */
-	else if (ldap_count_values(values)>1)
+	else if (l_count_values(values)>1)
 	 {
 		 *copy=strdup(values[0]);
 		 fprintf(stderr, "WARN: authldaplib: duplicate attribute %s for %s\n",
@@ -669,7 +825,7 @@ static void copy_value(LDAP *ld, LDAPMessage *entry, const char *attribut,
 		 *copy=NULL;
 	 }
   /* We accept only attribute with one value */
-	else if (ldap_count_values(values)!=1)
+	else if (l_count_values(values)!=1)
 	 {
 		 *copy=NULL;
 	 }
@@ -680,7 +836,7 @@ static void copy_value(LDAP *ld, LDAPMessage *entry, const char *attribut,
 #if DEBUG_LDAP
   DPRINTF("copy_value %s: %s",attribut,values[0]);
 #endif
-  ldap_value_free(values);
+  l_value_free(values);
 }
 
 static struct ldap_info my_ldap;
@@ -691,12 +847,12 @@ void authldapclose()
 {
 	if (my_ldap_fp)
 	{
-		ldap_unbind(my_ldap_fp);
+		l_unbind(my_ldap_fp);
 		my_ldap_fp=0;
 	}
 	if (bindp)
 	{
-		ldap_unbind(bindp);
+		l_unbind(bindp);
 		bindp=0;
 	}
 }
@@ -827,7 +983,9 @@ int     ldrc;
 		return (-1);
 	  }
 	if (my_ldap.protocol_version)
+	{
 		DPRINTF("selected ldap protocol version %d", my_ldap.protocol_version);
+	}
 #endif
 
 	if (my_ldap.tls && enable_tls_on(my_ldap_fp))
@@ -854,16 +1012,20 @@ int     ldrc;
 
   /* Bind to server */
   if (courier_authdebug_login_level >= 2)
+  {
 	  DPRINTF("binding to LDAP server as DN '%s', password '%s'",
 	  	my_ldap.binddn ? my_ldap.binddn : "<null>",
 	  	my_ldap.bindpw ? my_ldap.bindpw : "<null>");
+  }
   else
+  {
 	  DPRINTF("binding to LDAP server as DN '%s'",
 	  	my_ldap.binddn ? my_ldap.binddn : "<null>");
+  }
 
-  if (ldaperror(ldrc = ldap_simple_bind_s(my_ldap_fp,
-		(char *)my_ldap.binddn,
-		(char *)my_ldap.bindpw)) != LDAP_SUCCESS)
+  if (ldaperror(ldrc = l_simple_bind_s(my_ldap_fp,
+				       my_ldap.binddn,
+				       my_ldap.bindpw)) != LDAP_SUCCESS)
     {
     const char *s=ldap_err2string(ldrc);
 
@@ -915,10 +1077,29 @@ static int auth_ldap_do2(const char *service,
 
 static char *escape_str(const char *);
 
+static int auth_ldap_retry(const char *service,
+			   const char *user, const char *pass,
+			   int (*callback)(struct authinfo *, void *),
+			   void *arg, const char *newpass);
+
 static int auth_ldap_do(const char *service,
 			const char *user, const char *pass,
 			int (*callback)(struct authinfo *, void *),
                         void *arg, const char *newpass)
+{
+	int rc=auth_ldap_retry(service, user, pass, callback, arg, newpass);
+
+	if (rc > 0)
+		rc=auth_ldap_retry(service, user, pass, callback, arg,
+				   newpass);
+
+	return rc;
+}
+
+static int auth_ldap_retry(const char *service,
+			   const char *user, const char *pass,
+			   int (*callback)(struct authinfo *, void *),
+			   void *arg, const char *newpass)
 {
 	char *q;
 	int i;
@@ -982,14 +1163,14 @@ static int auth_ldap_do2(const char *service,
 		attributes[0]="handle";
 	attributes[1]=NULL;
 
-	if (ldaperror(ldap_search_st(my_ldap_fp,
-				     (char *)(my_ldap.emailmap_basedn[0] ?
-					      my_ldap.emailmap_basedn
-					      :my_ldap.basedn),
-				     LDAP_SCOPE_SUBTREE,
-				     srch, (char **)attributes, 0,
-				     &tv, &result))
-		      != LDAP_SUCCESS)
+	if (ldaperror(l_search_st(my_ldap_fp,
+				  (char *)(my_ldap.emailmap_basedn[0] ?
+					   my_ldap.emailmap_basedn
+					   :my_ldap.basedn),
+				  LDAP_SCOPE_SUBTREE,
+				  srch, (char **)attributes, 0,
+				  &tv, &result))
+	    != LDAP_SUCCESS)
 	{
 		free(srch);
 
@@ -1140,10 +1321,10 @@ static int auth_ldap_do3(const char *service,
 
 	my_ldap.attrlist[j]=0;
 
-	if (ldaperror(ldap_search_st(my_ldap_fp,
-				     (char *)my_ldap.basedn,LDAP_SCOPE_SUBTREE,
-				     filter, (char **)my_ldap.attrlist, 0,
-				     &timeout, &result))
+	if (ldaperror(l_search_st(my_ldap_fp,
+				  (char *)my_ldap.basedn,LDAP_SCOPE_SUBTREE,
+				  filter, (char **)my_ldap.attrlist, 0,
+				  &timeout, &result))
 		      != LDAP_SUCCESS)
 	{
 		DPRINTF("ldap_search_st() failed");
@@ -1170,7 +1351,7 @@ static int auth_ldap_do3(const char *service,
 
 	if (dn == NULL) 
 	{
-		ldap_perror(my_ldap_fp, "ldap_get_dn");
+		DPRINTF("ldap_get_dn failed");
 		return -1;
 	}
 
@@ -1178,7 +1359,7 @@ static int auth_ldap_do3(const char *service,
 	entry=ldap_first_entry(my_ldap_fp,result);
 	if (entry==NULL)
 	{
-		ldap_perror(my_ldap_fp,"ldap_first_entry");
+		DPRINTF("ldap_first_entry failed");
 		free(dn);
 		return -1;
 	}
@@ -1194,12 +1375,16 @@ static int auth_ldap_do3(const char *service,
 	BerElement *berptr = 0;
 
 		attr = ldap_first_attribute(my_ldap_fp, entry, &berptr);
-		if (attr) DPRINTF("raw ldap entry returned:");
+		if (attr)
+		{
+			DPRINTF("raw ldap entry returned:");
+		}
+
 		while (attr)
 		{
 		char **av, **ap;
 		
-			av = ldap_get_values(my_ldap_fp, entry, attr);
+			av = l_get_values(my_ldap_fp, entry, attr);
 			ap = av;
 			if (av)
 			{
@@ -1208,7 +1393,7 @@ static int auth_ldap_do3(const char *service,
 					DPRINTF("| %s: %s", attr, *ap);
 					ap++;
 				}
-				ldap_value_free(av);
+				l_value_free(av);
 			}
 			ldap_memfree(attr);
 			attr = ldap_next_attribute(my_ldap_fp, entry, berptr);
@@ -1396,7 +1581,7 @@ static int auth_ldap_do3(const char *service,
 				{
 				int ldrc;
 					DPRINTF("rebinding with DN '%s' to validate password", dn);
-					ldrc = ldap_simple_bind_s(bindp, dn, pass);
+					ldrc = l_simple_bind_s(bindp, dn, pass);
 					switch (ldrc)
 					{
 					case LDAP_SUCCESS:
@@ -1418,7 +1603,7 @@ static int auth_ldap_do3(const char *service,
 				   which didn't support rebinding on same conn */
 				if (rc > 0 || my_ldap.protocol_version == 2)
 				{
-					ldap_unbind(bindp);
+					l_unbind(bindp);
 					bindp=0;
 				}
 			}
@@ -1438,10 +1623,14 @@ static int auth_ldap_do3(const char *service,
 				if (strcmp(pass,auth.clearpasswd))
 				{
 					if (courier_authdebug_login_level >= 2)
+					{
 						DPRINTF("supplied password '%s' does not match clearpasswd '%s'",
 							pass, auth.clearpasswd);
+					}
 					else
+					{
 						DPRINTF("supplied password does not match clearpasswd");
+					}
 					rc= -1;
 				}
 			}
@@ -1511,7 +1700,7 @@ static int auth_ldap_do3(const char *service,
 			   than the search credentials (my_ldap_fp) for
 			   performing the password update. (May not always be
 			   true, ideally it would be configurable) */
-			ld_errno = ldap_modify_s(bindp? bindp:my_ldap_fp, dn, mods);
+			ld_errno = l_modify_s(bindp? bindp:my_ldap_fp, dn, mods);
 			if (ld_errno != LDAP_SUCCESS)
 			{
 				rc= -1;
@@ -1725,9 +1914,9 @@ void auth_ldap_enumerate( void(*cb_func)(const char *name,
 
 	DPRINTF("ldap_search: basedn='%s', filter='%s'",
 		my_ldap.basedn, my_ldap.enumerate_filter);
-	if ((msgid = ldap_search(my_ldap_fp, (char *)my_ldap.basedn,LDAP_SCOPE_SUBTREE,
-				     my_ldap.enumerate_filter,
-				     (char **)my_ldap.attrlist, 0)) < 0)
+	if ((msgid = l_search(my_ldap_fp, (char *)my_ldap.basedn,LDAP_SCOPE_SUBTREE,
+			      my_ldap.enumerate_filter,
+			      (char **)my_ldap.attrlist, 0)) < 0)
 	{
 		DPRINTF("ldap_search failed");
 		return;
@@ -1791,7 +1980,7 @@ void auth_ldap_enumerate( void(*cb_func)(const char *name,
 				continue;
 			}
 
-			n=ldap_count_values(names);
+			n=l_count_values(names);
 			if (n > 0)
 			{
 				const char *name = names[0] ? names[0] : "<null>";
@@ -1813,7 +2002,7 @@ void auth_ldap_enumerate( void(*cb_func)(const char *name,
 
 				if (attributes[2])
 				{
-					copy_value(my_ldap_fp, entry, attributes[1],
+					copy_value(my_ldap_fp, entry, attributes[2],
 						   &gid_s, name);
 				}
 
@@ -1891,7 +2080,7 @@ void auth_ldap_enumerate( void(*cb_func)(const char *name,
 				if (options)
 					free(options);
 			}
-			ldap_value_free(names);
+			l_value_free(names);
 
 			entry = ldap_next_entry(my_ldap_fp, entry);
 		}
