@@ -12,7 +12,7 @@
 #include	"rfc822.h"
 #include	"rfc2047.h"
 
-static const char rcsid[]="$Id: rfc2047.c,v 1.19 2004/10/21 00:10:51 mrsam Exp $";
+static const char rcsid[]="$Id: rfc2047.c,v 1.20 2006/01/22 03:33:24 mrsam Exp $";
 
 #define	RFC2047_ENCODE_FOLDLENGTH	76
 
@@ -525,7 +525,8 @@ static int encodebase64(const char *ptr, size_t len, const char *charset,
 			int (*func)(const char *, size_t, void *), void *arg,
 			size_t foldlen, size_t offset)
 {
-	unsigned char ibuf[3], obuf[4];
+	unsigned char ibuf[3];
+	char obuf[4];
 	int	i, rc;
 
 	while (len)
